;;;
;;; Universal Monitor 8080
;;;

	CPU	8080

TARGET	equ	"8080"

	INCLUDE	"config.inc"

	INCLUDE	"../common.inc"
	
;;;
;;; ROM area
;;;

	ORG	0000H
	DI
	JMP	CSTART

	ORG	0008H


	ORG	0010H
	JMP	CONIN

	ORG	0018H
	JMP	CONOUT

	ORG	0020H

	ORG	0028H

	;;
	;; Entry point
	;;

	ORG	ENTRY+0		; Cold start
E_CSTART:
	JMP	CSTART

	ORG	ENTRY+8		; Warm start
E_WSTART:
	JMP	WSTART

	ORG	ENTRY+16	; Console output
E_CONOUT:
	JMP	CONOUT

	ORG	ENTRY+24	; (Console) String output
E_STROUT:
	JMP	STROUT

	ORG	ENTRY+32	; Console input
E_CONIN:
	JMP	CONIN

	ORG	ENTRY+40	; Console status
E_CONST:
	JMP	CONST

	;;
	;;
	;;
	
	ORG	0100H
CSTART:
	LXI	SP,0FFFFH

	CALL	INIT

	LXI	HL,8000H
	SHLD	DSADDR
	SHLD	SADDR
	SHLD	GADDR
	MVI	A,'I'
	STA	HEXMOD
	XRA	A
	STA	PSPEC

	;; Opening message
	LXI	HL,OPNMSG
	CALL	STROUT

	;; CPU identification
	IF USE_IDENT
	MVI	A,7FH
	INR	A
	JPE	ID_Z80		; Z80, HD64180, Z280
	PUSH	PSW
	POP	BC
	MOV	A,C
	ANI	08H
	JNZ	ID_NEC		; NEC uPD8080A
	LXI	HL,0FFFFH
	INX	HL
	PUSH	PSW
	POP	BC
	MOV	A,C
	ANI	20H
	JNZ	ID_85		; i8085
	;; i8080 (uPD8080AF)
	LXI	HL,IM8080
	XRA	A
	JMP	IDE
	;; uPD8080A
ID_NEC:
	LXI	HL,IMNEC
	MVI	A,01H
	JMP	IDE
	;; i8085
ID_85:
	LXI	HL,IM8085
	MVI	A,02H
	JMP	IDE
	;; Z80
ID_Z80:
	LXI	HL,IMZ80
	MVI	A,10H
IDE:
	STA	PSPEC
	CALL	STROUT
	ENDIF

WSTART:
	LXI	HL,PROMPT
	CALL	STROUT
	CALL	GETLIN
	LXI	HL,INBUF
	CALL	SKIPSP
	CALL	UPPER
	ORA	A
	JZ	WSTART
	CPI	'D'
	JZ	DUMP
	CPI	'G'
	JZ	GO
	CPI	'S'
	JZ	SETM

	CPI	'L'
	JZ	LOADH
	CPI	'P'
	JZ	SAVEH

	CPI	'I'
	JZ	PIN
	CPI	'O'
	JZ	POUT
	
ERR:
	LXI	HL,ERRMSG
	CALL	STROUT
	JMP	WSTART

;;;
;;; Dump memory
;;; 
DUMP:
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX		; 1st arg.
	MOV	A,C
	ORA	A
	JNZ	DP0
	; No arg.
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JNZ	ERR
	LHLD	DSADDR
	LXI	BC,128
	DAD	BC
	SHLD	DEADDR
	JMP	DPM

DP0:	; 1st arg. found
	XCHG
	SHLD	DSADDR
	XCHG
	CALL	SKIPSP
	MOV	A,M
	CPI	','
	JZ	DP1
	ORA	A
	JNZ	ERR
	; No 2nd arg.
	LXI	HL,128
	DAD	DE
	SHLD	DEADDR
	JMP	DPM
DP1:
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	MOV	A,C
	ORA	A
	JZ	ERR
	MOV	A,M
	ORA	A
	JNZ	ERR
	INX	DE
	XCHG
	SHLD	DEADDR
	XCHG			; Might not needed
DPM:
	; DUMP main
	LHLD	DSADDR
	MVI	A,0F0H
	ANA	L
	MOV	L,A
	XRA	A
	STA	DSTATE
DPM0:
	PUSH	HL
	CALL	DPL
	POP	HL
	LXI	BC,16
	DAD	BC
	CALL	CONST
	JNZ	DPM1
	LDA	DSTATE
	CPI	2
	JC	DPM0
	LHLD	DEADDR
	SHLD	DSADDR
	JMP	WSTART
DPM1:
	CALL	CONIN
	SHLD	DSADDR
	JMP	WSTART

DPL:	; Dump line
	CALL	HEXOUT4
	PUSH	HL
	LXI	HL,DSEP0
	CALL	STROUT
	LXI	HL,INBUF
	SHLD	ASCPT
	POP	HL
	MVI	B,16
DPL0:
	CALL	DPB
	DCR	B
	JNZ	DPL0

	LXI	HL,DSEP1
	CALL	STROUT

	; Print ASCII area
	LXI	HL,INBUF
	MVI	B,16
DPL1:
	MOV	A,M
	INX	HL
	CPI	' '
	JC	DPL2
	CPI	7FH
	JNC	DPL2
	CALL	CONOUT
	JMP	DPL3
DPL2:
	MVI	A,'.'
	CALL	CONOUT
DPL3:
	DCR	B
	JNZ	DPL1
	JMP	CRLF

DPB:	; Dump byte
	MVI	A,' '
	CALL	CONOUT
	LDA	DSTATE
	ORA	A
	JNZ	DPB2
	; Dump state 0
	LDA	DSADDR		; Low byte
	CMP	L
	JNZ	DPB0
	LDA	DSADDR+1	; High byte
	CMP	H
	JZ	DPB1
DPB0:	; Still 0 or 2
	MVI	A,' '
	CALL	CONOUT
	CALL	CONOUT
	PUSH	HL
	LHLD	ASCPT
	MOV	M,A
	INX	HL
	SHLD	ASCPT
	POP	HL
	INX	HL
	RET
DPB1:	; Found start address
	MVI	A,1
	STA	DSTATE
DPB2:
	LDA	DSTATE
	CPI	1
	JNZ	DPB0
	; Dump state 1
	MOV	A,M
	PUSH	HL
	LHLD	ASCPT
	MOV	M,A
	INX	HL
	SHLD	ASCPT
	POP	HL
	CALL	HEXOUT2
	INX	HL
	LDA	DEADDR	; Low byte
	CMP	L
	RNZ
	LDA	DEADDR+1	; High byte
	CMP	H
	RNZ
	; Found end address
	MVI	A,2
	STA	DSTATE
	RET

GO:	; Go address
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	MOV	A,M
	ORA	A
	JNZ	ERR
	MOV	A,C
	ORA	A
	JZ	G0
	MOV	H,D
	MOV	L,E
	SHLD	GADDR
G0:
	LHLD	GADDR
	PCHL

SETM:	; Set memory
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JNZ	ERR
	XCHG		; EX DE,HL
	MOV	A,C
	ORA	A
	JNZ	SM0
	LHLD	SADDR
SM0:

SM1:
	CALL	HEXOUT4
	PUSH	HL
	LXI	HL,DSEP1
	CALL	STROUT
	POP	HL
	MOV	A,M
	PUSH	HL
	CALL	HEXOUT2
	MVI	A,' '
	CALL	CONOUT
	CALL	GETLIN
	LXI	HL,INBUF
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JNZ	SM2
	; Empty  (Increment address)
	POP	HL
	INX	HL
	SHLD	SADDR
	JMP	SM1
SM2:
	CPI	'-'
	JNZ	SM3
	; '-'  (Decrement address)
	POP	HL
	DCX	HL
	SHLD	SADDR
	JMP	SM1
SM3:
	CPI	'.'
	JNZ	SM4
	POP	HL
	SHLD	SADDR
	JMP	WSTART
SM4:
	CALL	RDHEX
	MOV	A,C
	ORA	A
	POP	HL
	JZ	ERR
	MOV	M,E
	INX	HL
	SHLD	SADDR
	JMP	SM1

;;;
;;; LOAD HEX file
;;;
LOADH:
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JNZ	ERR

	MOV	A,C
	ORA	A
	JNZ	LH0

	LXI	DE,0		; Offset
LH0:
	CALL	CONIN
	CALL	UPPER
	CPI	'S'
	JZ	LHS0
LH1:
	CPI	':'
	JZ	LHI0
LH2:
	;; Skip to EOL
	CPI	CR
	JZ	LH0
	CPI	LF
	JZ	LH0
LH3:
	CALL	CONIN
	JMP	LH2

LHI0:
	CALL	HEXIN
	MOV	C,A		; Checksum
	MOV	B,A		; Length

	CALL	HEXIN
	MOV	H,A		; Address H
	ADD	C
	MOV	C,A		; Checksum

	CALL	HEXIN
	MOV	L,A		; Address L
	ADD	C
	MOV	C,A		; Checksum

	;; Add offset
	DAD	DE

	CALL	HEXIN
	STA	RECTYP		; Record Type
	ADD	C
	MOV	C,A		; Checksum

	MOV	A,B
	ORA	A
	JZ	LHI3
LHI1:
	CALL	HEXIN
	PUSH	PSW
	ADD	C
	MOV	C,A		; Checksum

	LDA	RECTYP
	ORA	A
	JNZ	LHI20

	POP	PSW
	MOV	M,A
	INX	HL
	JMP	LHI2
LHI20:
	POP	PSW		; Drop
LHI2:
	DCR	B
	JNZ	LHI1
LHI3:
	CALL	HEXIN
	ADD	C
	JNZ	LHIE		; Checksum error
	LDA	RECTYP
	ORA	A
	JZ	LH3
	JMP	WSTART
LHIE:
	LXI	HL,IHEMSG
	CALL	STROUT
	JMP	WSTART

LHS0:
	CALL	CONIN
	STA	RECTYP		; Record type

	CALL	HEXIN
	MOV	B,A		; Length+3
	MOV	C,A		; Checksum

	CALL	HEXIN
	MOV	H,A		; Address H
	ADD	C
	MOV	C,A		; Checksum

	CALL	HEXIN
	MOV	L,A		; Address L
	ADD	C
	MOV	C,A		; Checksum

	DAD	DE

	DCR	B
	DCR	B
	DCR	B
	JZ	LHS3
LHS1:
	CALL	HEXIN
	PUSH	PSW
	ADD	C
	MOV	C,A		; Checksum

	LDA	RECTYP
	CPI	'1'
	JNZ	LHS2

	POP	PSW
	MOV	M,A
	INX	HL
	JMP	LHS20
LHS2:
	POP	PSW		; Drop
LHS20:
	DCR	B
	JNZ	LHS1
LHS3:
	CALL	HEXIN
	ADD	C
	CPI	0FFH
	JNZ	LHSE		; Checksum error

	LDA	RECTYP
	CPI	'7'
	JZ	WSTART
	CPI	'8'
	JZ	WSTART
	CPI	'9'
	JZ	WSTART
	JMP	LH3
LHSE:
	LXI	HL,SHEMSG
	CALL	STROUT
	JMP	WSTART
	
;;;
;;; SAVE HEX file
;;;
SAVEH:
	INX	HL
	MOV	A,M
	CALL	UPPER
	CPI	'I'
	JZ	SH0
	CPI	'S'
	JNZ	SH1
SH0:
	INX	HL
	STA	HEXMOD
SH1:
	CALL	SKIPSP
	CALL	RDHEX
	MOV	A,C
	ORA	A
	JZ	SHE
	XCHG
	SHLD	ASCPT		; (ASCPT) = Start address
	XCHG
	CALL	SKIPSP
	MOV	A,M
	CPI	','
	JNZ	SHE
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	MOV	A,C
	ORA	A
	JZ	SHE
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JZ	SH2
SHE:
	JMP	ERR

SH2:
	LHLD	ASCPT
	INX	DE
	MOV	A,E
	SUB	L
	STA	COUNT
	MOV	A,D
	SBB	H
	STA	COUNT+1		; (COUNT) = Length
SH3:
	CALL	SHL
	LDA	COUNT
	MOV	B,A
	LDA	COUNT+1
	ORA	B
	JNZ	SH3
	
	LDA	HEXMOD
	CPI	'I'
	JNZ	SH4
	;; End record for Intel HEX
	LXI	HL,IHEXER
	CALL	STROUT
	JMP	WSTART
SH4:
	;; End record for Motorola S record
	LXI	HL,SRECER
	CALL	STROUT
	JMP	WSTART

SHL:
	MVI	B,16
	LDA	COUNT+1		; Length H
	ORA	A
	JNZ	SHL0
	LDA	COUNT		; Length L
	CMP	B
	JNC	SHL0
	MOV	B,A
SHL0:
	LDA	COUNT
	SUB	B
	STA	COUNT
	LDA	COUNT+1
	SBI	0
	STA	COUNT+1

	LDA	HEXMOD
	CPI	'I'
	JNZ	SHLS

	;; Intel HEX
	MVI	A,':'
	CALL	CONOUT

	MOV	A,B
	CALL	HEXOUT2		; Length
	MOV	A,B
	MOV	C,A		; Checksum

	MOV	A,H
	CALL	HEXOUT2		; Address H
	MOV	A,H
	ADD	C
	MOV	C,A

	MOV	A,L
	CALL	HEXOUT2		; Address L
	MOV	A,L
	ADD	C
	MOV	C,A

	XRA	A
	CALL	HEXOUT2
SHLI0:
	MOV	A,M
	PUSH	PSW
	CALL	HEXOUT2
	POP	PSW
	ADD	C
	MOV	C,A

	INX	HL
	DCR	B
	JNZ	SHLI0

	MOV	A,C
	CMA
	INR	A
	CALL	HEXOUT2
	JMP	CRLF

SHLS:
	;; Motorola S record
	MVI	A,'S'
	CALL	CONOUT
	MVI	A,'1'
	CALL	CONOUT

	MOV	A,B
	ADI	2+1		; DataLength + Addr(2) + Sum(1)
	MOV	C,A
	CALL	HEXOUT2

	MOV	A,H
	CALL	HEXOUT2		; Address H
	MOV	A,H
	ADD	C
	MOV	C,A

	MOV	A,L
	CALL	HEXOUT2		; Address L
	MOV	A,L
	ADD	C
	MOV	C,A
SHLS0:
	MOV	A,M
	PUSH	PSW
	CALL	HEXOUT2		; Data
	POP	PSW
	ADD	C
	MOV	C,A

	INX	HL
	DCR	B
	JNZ	SHLS0

	MOV	A,C
	CMA
	CALL	HEXOUT2		; Checksum
	JMP	CRLF

;;;
;;; Port in
;;;
PIN:
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	MOV	A,C
	ORA	A
	JZ	ERR		; Port addr. missing
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JNZ	ERR

	LXI	HL,INBUF
	MVI	M,0DBH		; OPCODE of IN
	INX	HL
	MOV	M,E		; Port address
	INX	HL
	MVI	M,0C9H		; OPCODE of RET
	CALL	INBUF
	CALL	HEXOUT2
	CALL	CRLF
	JMP	WSTART

;;;
;;; Port out
;;;
POUT:
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	MOV	A,C
	ORA	A
	JZ	ERR		; Port addr. missing
	MOV	B,E		; Port address
	CALL	SKIPSP
	MOV	A,M
	CPI	','
	JNZ	ERR
	INX	HL
	CALL	SKIPSP
	CALL	RDHEX
	MOV	A,C
	ORA	A
	JZ	ERR		; Data missing
	CALL	SKIPSP
	MOV	A,M
	ORA	A
	JNZ	ERR

	LXI	HL,INBUF
	MVI	M,0D3H		; OPCODE of OUT
	INX	HL
	MOV	M,B		; Port address
	INX	HL
	MVI	M,0C9H		; OPCODE of RET
	MOV	A,E		; Data
	CALL	INBUF
	JMP	WSTART
	
;;;
;;; Other support routines
;;;

STROUT:
	MOV	A,M
	ANA	A
	RZ
	CALL	CONOUT
	INX	HL
	JMP	STROUT

HEXOUT4:
	MOV	A,H
	CALL	HEXOUT2
	MOV	A,L
HEXOUT2:
	PUSH	PSW
	RAR
	RAR
	RAR
	RAR
	CALL	HEXOUT1
	POP	PSW
HEXOUT1:
	ANI	0FH
	ADI	'0'
	CPI	'9'+1
	JC	CONOUT
	ADI	'A'-'9'-1
	JMP	CONOUT

HEXIN:
	XRA	A
	CALL	HI0
	RLC
	RLC
	RLC
	RLC
HI0:
	PUSH	BC
	MOV	C,A
	CALL	CONIN
	CALL	UPPER
	CPI	'0'
	JC	HIR
	CPI	'9'+1
	JC	HI1
	CPI	'A'
	JC	HIR
	CPI	'F'+1
	JNC	HIR
	SUI	'A'-'9'-1
HI1:
	SUI	'0'
	ORA	C
HIR:
	POP	BC
	RET
	
CRLF:
	MVI	A,CR
	CALL	CONOUT
	MVI	A,LF
	JMP	CONOUT

GETLIN:
	LXI	HL,INBUF
	MVI	B,0
GL0:
	CALL	CONIN
	CPI	CR
	JZ	GLE
	CPI	LF
	JZ	GLE
	CPI	BS
	JZ	GLB
	CPI	DEL
	JZ	GLB
	CPI	' '
	JC	GL0
	CPI	80H
	JNC	GL0
	MOV	C,A
	MOV	A,B
	CPI	BUFLEN-1
	JNC	GL0		; Too long
	INR	B
	MOV	A,C
	CALL	CONOUT
	MOV	M,A
	INX	HL
	JMP	GL0
GLB:
	MOV	A,B
	ANA	A
	JZ	GL0
	DCR	B
	DCX	HL
	MVI	A,08H
	CALL	CONOUT
	MVI	A,' '
	CALL	CONOUT
	MVI	A,08H
	CALL	CONOUT
	JMP	GL0
GLE:
	CALL	CRLF
	MVI	M,00H
	RET

SKIPSP:
	MOV	A,M
	CPI	' '
	RNZ
	INX	HL
	JMP	SKIPSP

UPPER:
	CPI	'a'
	RC
	CPI	'z'+1
	RNC
	ADI	'A'-'a'
	RET

RDHEX:
	MVI	C,0
	LXI	DE,0
RH0:
	MOV	A,M
	CALL	UPPER
	CPI	'0'
	JC	RHE
	CPI	'9'+1
	JC	RH1
	CPI	'A'
	JC	RHE
	CPI	'F'+1
	JNC	RHE
	SUI	'A'-'9'-1
RH1:
	SUI	'0'
	RAL
	RAL
	RAL
	RAL
	RAL
	PUSH	PSW
	MOV	A,E
	RAL
	MOV	E,A
	MOV	A,D
	RAL
	MOV	D,A
	POP	PSW
	RAL
	PUSH	PSW
	MOV	A,E
	RAL
	MOV	E,A
	MOV	A,D
	RAL
	MOV	D,A
	POP	PSW
	RAL
	PUSH	PSW
	MOV	A,E
	RAL
	MOV	E,A
	MOV	A,D
	RAL
	MOV	D,A
	POP	PSW
	RAL
	MOV	A,E
	RAL
	MOV	E,A
	MOV	A,D
	RAL
	MOV	D,A
	INX	HL
	INR	C
	JMP	RH0
RHE:
	RET


DMYINT:
	EI
	RET

OPNMSG:
	DB	CR,LF,"Universal Monitor 8080",CR,LF,00H

PROMPT:
	DB	"] ",00H

IHEMSG:
	DB	"Error ihex",CR,LF,00H

SHEMSG:
	DB	"Error srec",CR,LF,00H

ERRMSG:
	DB	"Error",CR,LF,00H

DSEP0:
	DB	" :",00H
DSEP1:
	DB	" : ",00H
IHEXER:
        DB	":00000001FF",CR,LF,00H
SRECER:
        DB	"S9030000FC",CR,LF,00H

	IF USE_IDENT
IM8080:
	DB	"i8080",CR,LF,00H
IMNEC:
	DB	"uPD8080A",CR,LF,00H
IM8085:
	DB	"i8085",CR,LF,00H
IMZ80:
	DB	"Z80",CR,LF,00H
	ENDIF

	IF USE_DEV_8251
	INCLUDE	"dev/unimon_dev_8251.asm"
	ENDIF
	
;;;
;;; RAM area
;;;

	;; 
	;; Work area
	;;
	
	ORG	WORK_B

INBUF:	DS	BUFLEN		; Line input buffer
DSADDR:	DS	2		; Dump start address
DEADDR:	DS	2		; Dump end address
DSTATE:	DS	1		; Dump state
GADDR:	DS	2		; Go address
SADDR:	DS	2		; Set address
HEXMOD:	DS	1		; HEX file mode
PSPEC:	DS	1		; Processor spec.

ASCPT:	DS	2		; ASC buffer pointer (DUMP), address pointer (PUNCH)
COUNT:	DS	2
RECTYP:	DS	1		; Record Type
	
	END
